# -*- coding: utf-8 -*-
'''*****************************************************************************
***  This file is part of HgWeb Manager.                                     ***
***                                                                          ***
***  Copyright (C) 2014                                                      ***
***  CodeGNU Solutions <Licensing _AT_ CodeGNU _DOT_ com>                    ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*****************************************************************************'''


"""
IMPORTS
"""
from hwm import settings


def create(repository, ignoreGroups, skipCommon):
	""" Creates a new repository under HgWeb.

	@pre The repository to create cannot already exist.

	@param[in] repository		The targeted repository to create.
	@param[in] ignoreGroups		The list of all ignore groups to add.
	@param[in] skipCommon		A flag that when true indicates the common group
								should not be added by default.

	@post The file system now contains a new directory in the repository path
	that is the newly generated repository. The new repository is registered
	in HgWeb to display and is secured to prevent accidental modification.

	@return When the creation is successful gives true, else-wise false.
	"""
	import sys
	import os
	import shutil
	import stat

	with RepositoryManagerLock():
		print "Adding repository: %s" % repository.StorageName
		if __doesRepositoryStorageNameExist(repository):
			print >>sys.stderr, "The repository storage name '%s' already exists." % repository.StorageName
			return False

		if __doesRepositoryDisplayNameExist(repository):
			print >>sys.stderr, "The display name '%s' already exists." % repository.DisplayName
			return False

		# Create the requested repository.
		if __addRepository(repository):
			print "Initialised repository directory."
			if not __create_ignores(repository, ignoreGroups, skipCommon):
				__removeRepository(repository)
				return False
		else:
			print >>sys.stderr, "Failed to create repository directory."
			return False

		# Fill out the details of about the repository.
		if __writeRepositoryDetails(repository):
			print "Successfully wrote display details."
		else:
			print >>sys.stderr, "Failed to write display details."
			os.chmod(settings.RepositoryPath + os.sep + repository.StorageName + os.sep + '.hg' + os.sep + "hgrc", stat.S_IWRITE)
			shutil.rmtree(settings.RepositoryPath + os.sep + repository.StorageName)
			return False

		if __registerRepository(repository):
			print 'Repository added.'
			return True
		else:
			os.chmod(settings.RepositoryPath + os.sep + repository.StorageName + os.sep + '.hg' + os.sep + "hgrc", stat.S_IWRITE)
			shutil.rmtree(settings.RepositoryPath + os.sep + repository.StorageName)
			return False


def register(repository):
	""" Connects an existing repository to the mercurial web registry.

	@pre The repository to register must already exist.

	@param[in] repository	The targeted repository to register.

	@post The new repository is registered in HgWeb to display and is
	secured to prevent accidental modification.

	@return When the registration is successful gives true, else-wise false.
	"""
	with RepositoryManagerLock():
		print "Registering the repository: %s" % repository.StorageName
		if not __doesRepositoryStorageNameExist(repository):
			print >>sys.stderr, "The repository '%s' does not exists." % repository.StorageName
			return False

		# Add the new repository to the web registry.
		if __registerRepository(repository):
			print "Successfully registered repository."
			return True
		else:
			print >>sys.stderr, "Failed to register repository."
			return False


def modify(newRepository, currentStorageName, ignoreGroups, skipCommon):
	""" Changes an existing repository in HgWeb.

	@pre The repository to modify must already exist.

	@param[in] newRepository		The new values for the repository to
									modify.
	@param[in] currentStorageName	The current storage name of the
									repository to modify.
	@param[in] ignoreGroups			The list of all ignore groups to add.
	@param[in] skipCommon			A flag that when true indicates the
									common group should not be added by
									default.

	@post The repository is changed with the requested modifications.

	@return When the modification is successful gives true, else-wise false.
	"""
	import sys
	from manrepo import Repository

	with RepositoryManagerLock():
		print "Modifying repository: %s" % currentStorageName
		oldRepository = Repository(currentStorageName)

		if not __doesRepositoryStorageNameExist(oldRepository):
			print >>sys.stderr, "The repository '%s' was not found." % oldRepository.StorageName
			return False

		# Remove the current repository from the web registry.
		if not __unregisterRepository(oldRepository):
			print >>sys.stderr, "Failed to unregister repository."
			return False

		if __doesRepositoryDisplayNameExist(newRepository):
			print >>sys.stderr, "The display name '%s' already exists." % newRepository.DisplayName
			if not __registerRepository(oldRepository):
				print >>sys.stderr, "Failed to register repository."
			return False

		# Update renamed repositories
		if not oldRepository == newRepository:
			if __renameStorage(oldRepository, newRepository.StorageName):
				print "Renamed repository %s to %s." % (currentStorageName, newRepository.StorageName)
			else:
				print >>sys.stderr, "Failed to rename repository."
				if __registerRepository(oldRepository):
					print "Successfully rolled backed action."
				else:
					print >>sys.stderr, "Failed to roll back action."
				return False

		# Fill out the details of about the new repository.
		if __writeRepositoryDetails(newRepository):
			print "Successfully wrote new details."
		else:
			print >>sys.stderr, "Failed to write new details."
			if __registerRepository(oldRepository):
				print "Successfully rolled backed action."
			else:
				print >>sys.stderr, "Failed to roll back action."
			return False

		# Update ignore file if requested.
		if ignoreGroups is not None:
			if __create_ignores(newRepository, ignoreGroups, skipCommon):
				print "Successfully modified ignore file."
			else:
				print >>sys.stderr, "Failed to modify ignore file."
				return False

		# Add the new repository to the web registry.
		if __registerRepository(newRepository):
			print 'Repository modified.'
			return True
		else:
			print >>sys.stderr, "Failed to register repository."
			return False


def delete(repository):
	""" Removes an existing repository.

	@pre The repository to remove must already exist.

	@param[in] repository	The targeted repository to delete.

	@post The file system no longer contains a directory in the repository
	path that is the supplied repository. The repository is no longer
	registered in HgWeb.

	@return When the deletion is successful gives true, else-wise false.
	"""
	with RepositoryManagerLock():
		if not __doesRepositoryStorageNameExist(repository):
			print >>sys.stderr, "The repository %s was not found." % repository.StorageName
			return False

		print "Removing repository: %s" % repository.StorageName
		if __unregisterRepository(repository):
			print "Successfully unregistered repository."
		else:
			print >>sys.stderr, "Failed to unregister repository."

		if __removeRepository(repository):
			print "Repository removed."
			return True
		else:
			print >>sys.stderr, "Failed to delete the repository."
			return False
		print >>sys.stderr, "The HgWeb configuration file does not have a paths section!"


def list():
	""" Outputs a listing of all the known managed repositories

	@return When the output generation is successful gives true, else-wise false.
	"""
	import manrepo

	print "Managed Repositories:"
	try:
		repoCol = manrepo.ManagedCollection()
		for repo in repoCol.Repositories:
			print "\t%s (%s)" % (repo.StorageName, repo.DisplayName)
		return True
	except Exception as e:
		import sys

		print >>sys.stderr, "Listing error({0}): {1}".format(e.errno, e.strerror)
		return False

# Functions for fulfilling repository management.


def __create_ignores(repo, ignoreGroups, skipCommon):
	""" Generates and commits an ignore file to an existing repository.

	@pre The repository to create an ignore file must already exist. The list
	of ignore patterns to use is expected to be in an XML file named
	"ignores.xml" and be in the hwm-ignore schema.

	@param[in] repo				The targeted repository for the ignore file.
	@param[in] ignoreGroups		The list of all ignore groups to add.
	@param[in] skipCommon		A flag that when true indicates the common group
								should not be added by default.

	@post The managed repository now has an ignore file committed by the
	HgWeb manager.

	@return When the creation is successful gives true, else-wise false.
	"""
	import os
	import ignorepo
	import sys
	import xml

	# Extract Ignore Collection
	if ignoreGroups is None:
		ignoreGroups = []
	ignores = {}
	try:
		ignores = ignorepo.extract('ignores.xml')
	except IOError as e:
		print >>sys.stderr, "Failed to locate ignores, error({0}): {1}".format(e.errno, e.strerror)
		return False
	except xml.parsers.expat.ExpatError as e:
		print >>sys.stderr, "Failed to extract ignores, error: %s" % e.message
		return False

	if skipCommon:
		del ignores['Common']
	elif 'common' not in [ignoreGroup.lower() for ignoreGroup in ignoreGroups]:
		ignoreGroups.append('Common')

	if '+' not in ignoreGroups:
		for ignoreGroup in ignoreGroups:
			if not ignoreGroup.lower() in [ignore.lower() for ignore in ignores]:
				print >>sys.stderr, "The list of ignores does not have the '%s' group." % ignoreGroup
				return False

	ignoreFile = settings.RepositoryPath + os.sep + repo.StorageName + os.sep + '.hgignore'
	wasCreated = not os.path.isfile(ignoreFile)
	if wasCreated:
		print "Creating ignore file ..."
	else:
		print "Rewriting ignore file ..."

	# Write Ignore File
	with open(ignoreFile, 'w') as hgIgnore:
		for groupName, groupPatterns in ignores.items():
			if '+' not in ignoreGroups and not groupName.lower() in [ignoreGroup.lower() for ignoreGroup in ignoreGroups]:
				continue
			hgIgnore.write("#\n#\tIgnore Group: %s\n#" % groupName)
			# Reverse the sorting order to put the regular expressions first.
			groupPatterns.sort(key=lambda p: p.Filter, reverse=True)
			filteringOn = None
			for pattern in groupPatterns:
				if pattern.Filter == filteringOn:
					hgIgnore.write("\n%s\t\t#  %s" % (pattern.Value, pattern.Description))
				elif filteringOn is None:
					hgIgnore.write("\nsyntax: %s\n%s\t\t#  %s" % (ignorepo.ignore_syntax_name(pattern.Filter), pattern.Value, pattern.Description))
				else:
					hgIgnore.write("\n\nsyntax: %s\n%s\t\t#  %s" % (ignorepo.ignore_syntax_name(pattern.Filter), pattern.Value, pattern.Description))
				filteringOn = pattern.Filter
			hgIgnore.write('\n\n')

	if wasCreated:
		if __addFile(repo, '.hgignore'):
			print "Successfully added generated ignores."
		else:
			print >>sys.stderr, "Failed to add generated ignores."
			return False

	if __commitChanges(repo, 'Created generated ignores filter.'):
		print "Successfully committed generated ignores."
	else:
		print >>sys.stderr, "Failed to commit generated ignores."
		return False
	return True


def __commitChanges(repo, message):
	""" Commits all changes in a managed Mercurial repository as the HgWeb
	manager.

	@pre The requested repository must already exist.

	@param[in] repo		The targeted repository for which to commit.
	@param[in] message	The text to use as a commit message.

	@post Upon any error the standard error buffer contains an error message.

	@return When the process is successful gives true, else-wise false.
	"""
	import subprocess
	import sys
	import os

	try:
		statusCode = subprocess.call([settings.HgCommand, '--repository', settings.RepositoryPath + os.sep + repo.StorageName, 'commit', '--message', message, '--user', settings.HgUser + ' - HgWeb Manager'])
	except OSError as e:
		print >>sys.stderr, "Commit Error({0}): {1}".format(e.errno, e.strerror)
		return False
	# Zero being a successful commit and one being nothing changed.
	return ((0 == statusCode) or (1 == statusCode))


def __addFile(repo, fileName):
	""" Marks a file as added in a managed Mercurial repository.

	@pre The requested repository must already exist.

	@param[in] repo			The targeted repository for which to add the file.
	@param[in] fileName		The directory path and the file name to add
							relative to the base of the repository.

	@post Upon any error the standard error buffer contains an error message.

	@return When the process is successful gives true, else-wise false.
	"""
	import subprocess
	import sys
	import os

	try:
		statusCode = subprocess.call([settings.HgCommand, '--repository', settings.RepositoryPath + os.sep + repo.StorageName, 'add', settings.RepositoryPath + os.sep + repo.StorageName + os.sep + fileName])
	except OSError as e:
		print >>sys.stderr, "Add Error({0}): {1}".format(e.errno, e.strerror)
		return False
	return (0 == statusCode)


def __renameStorage(oldRepo, newName):
	""" Changes the storage name of an existing repository.

	@param[in] oldRepo		The existing repository to whose storage name is
							expected to be changed.
	@param[in] newName		The new storage name for the repository.

	@return Gives true when the storage name is successfully renamed,
	else-wise gives false.
	"""
	import os
	import sys
	import shutil

	try:
		if oldRepo.StorageName == newName:
			return True
		elif os.path.isdir(settings.RepositoryPath + os.sep + newName):
			print >>sys.stderr, "Cannot rename the existing repository '%s' because the name '%s' already exists." % (oldRepo.StorageName, newName)
			return False
		else:
			shutil.move(settings.RepositoryPath + os.sep + oldRepo.StorageName, settings.RepositoryPath + os.sep + newName)
	except OSError as e:
		print >>sys.stderr, "Rename error({0}): {1}".format(e.errno, e.strerror)
		return False
	return True


def __doesRepositoryStorageNameExist(repo):
	""" Checks to see if a repository storage name exists.

	@param[in] repo		The targeted repository for which to check.

	@return Gives true when it exists, false else-wise.
	"""
	import os

	if (repo is None) or (repo.StorageName is None):
		return False
	elif os.path.isdir(settings.RepositoryPath + os.sep + repo.StorageName + os.sep + '.hg'):
		return True
	else:
		return False


def __doesRepositoryDisplayNameExist(repo):
	""" Checks to see if a repository display name exists.

	@param[in] repo		The targeted repository for which to check.

	@return Gives true when it exists, false else-wise.
	"""
	from manrepo import ManagedCollection

	if (repo is None) or (repo.StorageName is None) or (repo.DisplayName is None):
		return False

	repoCol = ManagedCollection()
	for testRepo in repoCol.Repositories:
		if (testRepo.DisplayName is not None) and (not testRepo == repo) and (testRepo.DisplayName.lower() == repo.DisplayName.lower()):
			return True
	return False


def __unregisterRepository(repo):
	""" Removes the registration of the supplied repository.

	@param[in] repo		The targeted repository for which to unregister.

	@post The HgWeb configuration file does not contain the repository details
	and is returned to a read-only state. Upon any error the standard error
	buffer contains an error message.

	@return When the process is successful gives true, else-wise false.
	"""
	from ConfigParser import SafeConfigParser
	import os
	import sys

	parser = SafeConfigParser()
	if not parser.read(settings.HgWebPath + os.sep + 'hgweb.config'):
		print >>sys.stderr, "Failed to locate HgWeb configuration file."
		return False

	if parser.has_section('paths'):
		if parser.has_option('paths', repo.StorageName):
			print "Unregistering repository ..."
			parser.remove_option('paths', repo.StorageName)
			with open(settings.HgWebPath + os.sep + 'hgweb.config', 'wb') as configfile:
				parser.write(configfile)
			return True
		else:
			print "The repository already is unregistered."
			return True
	else:
		print >>sys.stderr, "The HgWeb configuration file does not have a paths section!"
		return False


def __removeRepository(repo):
	""" Removes an existing Mercurial repository.

	@pre The requested repository to delete must already exist.

	@param[in] repo		The targeted repository for which to delete.

	@post Upon any error the standard error buffer contains an error message.

	@return When the process is successful gives true, else-wise false.
	"""
	import shutil
	import os
	import stat

	try:
		os.chmod(settings.RepositoryPath + os.sep + repo.StorageName + os.sep + '.hg' + os.sep + "hgrc", stat.S_IWRITE)
	except OSError as e:
		# If it's not the "No such file or directory", notify the user.
		if not 2 == e.errno:
			import sys

			print >>sys.stderr, "Repository removal error({0}): {1}".format(e.errno, e.strerror)
			return False
	shutil.rmtree(settings.RepositoryPath + os.sep + repo.StorageName)
	return True


def __addRepository(repo):
	""" Initialises a new Mercurial repository.

	@pre The requested repository to create cannot already exist.

	@param[in] repo	The targeted repository for which to create.

	@post Upon any error the standard error buffer contains an error message.

	@return When the process is successful gives true, else-wise false.
	"""
	import subprocess
	import sys
	import os

	if repo.StorageName is None:
		print >>sys.stderr, 'Cannot create a managed repository without a storage name.'
		return False

	try:
		statusCode = subprocess.call([settings.HgCommand, 'init', settings.RepositoryPath + os.sep + repo.StorageName])
	except OSError as e:
		print >>sys.stderr, "Initialisation error({0}): {1}".format(e.errno, e.strerror)
		return False
	return (0 == statusCode)


def __writeRepositoryDetails(repo):
	""" Writes the details of a repository to the repository's
	configuration file.

	@param[in] repo		The targeted repository for which to write the
						configuration details.

	@post The HgWeb configuration file contains the new repository details and
	is returned to a read-only state.

	@return When the process is successful gives true, else-wise false.
	"""
	import os
	import stat
	import sys

	print "Saving display details ..."
	config = settings.RepositoryPath + os.sep + repo.StorageName + os.sep + '.hg' + os.sep + 'hgrc'
	if os.path.exists(config):
		hgrcMode = os.stat(config)[stat.ST_MODE]
		os.chmod(config, hgrcMode | stat.S_IWRITE)
	else:
		hgrcMode = stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH

	try:
		repo.save()
	except IOError as e:
		print >>sys.stderr, "Failed to save details, error({0}): {1}".format(e.errno, e.strerror)
		return False
	# Ensure the repository details isn't accidentally modified.
	os.chmod(config, hgrcMode & (~stat.S_IWUSR & ~stat.S_IWGRP & ~stat.S_IWOTH))

	# Ensure the repository isn't accidentally deleted, moved, or blocked in Windows.
	if "win32" == sys.platform:
		if __guardWindowsRepository(repo):
			print 'Secured repository store in Windows.'
		else:
			print >>sys.stderr, 'Failed to secure repository store in Windows.'
			return False
	return True


def __guardWindowsRepository(repo):
	""" Guards the supplied repository for a repository on a Windows system.

	@param[in] repo	The targeted repository for which to secure.

	@return When the process is successful gives true, else-wise false.
	"""
	import subprocess
	import os

	try:
		print 'Securing for Windows ...'
		status = subprocess.call(["attrib.exe", '+H', '+I', '+S', settings.RepositoryPath + os.sep + repo.StorageName + os.sep + '.hg'])
	except OSError as e:
		import sys

		print >>sys.stderr, "Windows Guard error({0}): {1}".format(e.errno, e.strerror)
		return False
	return (0 == status)


def __registerRepository(repo):
	""" Adds the registration details of the repository.

	@param[in] repo		The targeted repository for which to register in HgWeb.

	@post The HgWeb configuration file contains the new repository details and
	is returned to a read-only state. Upon any error the standard error
	buffer contains an error message.

	@return When the process is successful gives true, else-wise false.
	"""
	from ConfigParser import SafeConfigParser
	import os
	import sys

	parser = SafeConfigParser()
	if not parser.read(settings.HgWebPath + os.sep + 'hgweb.config'):
		print >>sys.stderr, "Failed to locate HgWeb configuration file."
		return False

	if parser.has_section('paths'):
		if parser.has_option('paths', repo.StorageName):
			print >>sys.stderr, "The repository already is registered."
			return False
	else:
		parser.add_section('paths')
	print >>sys.stderr, "Registering repository ..."
	parser.set('paths', repo.StorageName, settings.RepositoryPath + os.sep + repo.StorageName)
	with open(settings.HgWebPath + os.sep + 'hgweb.config', 'wb') as configfile:
		parser.write(configfile)
	return True
